import numpy as np
import pandas as pd

def sir_step (S, I , R, beta, gamma, N):
    Sn = (-beta *S*I) + S
    In = (beta*S*I - gamma * I) + I
    Rn = gamma*I + R
    Sn, Rn, In = (0 if x <0 else x for x in [Sn, Rn, In])

    scale = N / (Sn + In + Rn)
    return (x*scale for x in [Sn, Rn, In])


def chime_sir(S, I, R, beta, gamma, n_days, beta_decay = None):
    N = sum([S, I, R])
    s, i, r = ([x] for x in  [S, I, R])

    for _ in range(n_days):
        S, I, R = sir_step(S, I, R, beta, gamma, N)
        beta = beta*(1-beta_decay) if beta_decay is not None else beta
        s += [S]
        i += [I]
        r += [R]

    return (np.array(x) for x in [s,i,r])


def generate_pars(S, infect_0, curr_hosp, hosp_rate, t_double,
                  contact_rate, hosp_share, hos_los, icu_los,
                  vent_los, R, t_rec, vent_rate, icu_rate):
    out = {}
    out['S'] = S
    out['infection_known'] = infect_0
    out['hosp_rate'] = hosp_rate
    out['vent_rate'] = vent_rate
    out['icu_rate'] = icu_rate
    out['hosp_los'] = hos_los
    out['icu_los'] = icu_los
    out['vent_los'] = vent_los
    out['hosp_share'] = hosp_share
    infect_total = curr_hosp / hosp_share / hosp_rate
    out['I'] = infect_total
    out['detect_prob'] = infect_0 / infect_total
    out['R'] = R
    out['growth_intrinsic'] = 2**(1/t_double)  - 1
    out['t_rec'] = t_rec
    out['gamma'] = 1 / t_rec
    out['contact_rate'] = contact_rate
    out['beta'] = ((out['growth_intrinsic'] + out['gamma']) / S) * (1-contact_rate)
    out['r_t'] = out['beta'] / out['gamma'] * S
    out['r_naught'] = out['r_t'] / (1-contact_rate)
    out['t_double_base'] = t_double
    out['t_double_true'] = 1/np.log2(out['beta']*S - out['gamma'] + 1)
    return out


def chime(S, infect_0, curr_hosp, hosp_rate=0.05, t_double=6,
          contact_rate=0, hosp_share = 1., hos_los=7, icu_los=9,
          vent_los=10, R = 0, t_rec = 14, beta_decay = None, n_days = 60,
          vent_rate=0.01, icu_rate=0.02):
    '''
    chime: SIR model for ventilators, ICU , and hospitilaizations.
    -----------------------------
    parameters:
        S: population size
        infect_0: number of confirmed infections
        curr_hosp: number of currently hospitalized patients
        hosp_rate: hospitalization_rate = 0.05
        t_double: time to double without distancing = 6
        contact_rate: reduction in contact due to distancing = 0
        hosp_share: proportion of the population represented by hospitalization data = 1
        hos_los: how long people stay in the hospital = 7
        icu_los : how long people stay in the ice = 9
        vent_los : how long people stay with ventilators = 10
        R : number recovered present data = 0
        t_rec : number of days to recover = 14
        betay_decay : beta decay = None
        n_days: number of days ahead to look = 60
        vent_rate: rate of people who need ventilators = 0.01
        icu_rate: rate of people who need icu = 0.02
    '''

    pars = generate_pars(S, infect_0, curr_hosp, hosp_rate, t_double,
                  contact_rate, hosp_share, hos_los, icu_los,
                  vent_los, R, t_rec, vent_rate, icu_rate)

    s, i, r = chime_sir(pars['S'], pars['I'], pars['R'],
                        pars['beta'], pars['gamma'], n_days, beta_decay)

    hosp, vent, icu = (pars[x]*i*pars['hosp_share'] for x in ['hosp_rate', 'vent_rate','icu_rate'])
    # something is wrong in general, but will be up and running by bedtime
    days = np.arange(0, n_days+1)
    data = dict(zip(['day','hosp','icu','vent'], [days, hosp, icu, vent]))
    proj = pd.DataFrame.from_dict(data)
    proj_admits = proj.iloc[:-1,:] - proj.shift(1)
    proj_admits[proj_admits <0 ] = 0
    plot_days = n_days - 10
    proj_admits["day"] = range(proj_admits.shape[0])
    los_dict = {k[:-4]: pars[k] for k in ['hosp_los', 'icu_los', 'vent_los']}
    census_dict = {}
    for k, los in los_dict.items():
        census = (
            proj_admits.cumsum().iloc[:-los,:] - proj_admits.cumsum().shift(los).fillna(0)
        ).apply(np.ceil)
        census_dict[k] = census[k]
    census_df = pd.DataFrame(census_dict)
    census_df["day"] = census_df.index
    census_df = census_df[["day", "hosp", "icu", "vent"]]

    census_table = census_df[np.mod(census_df.index, 7) == 0].copy()
    census_table.index = range(census_table.shape[0])
    census_table.loc[0, :] = 0
    census_table = census_table.dropna().astype(int)
    return data, proj_admits, census_table


dtx, admits, census = chime(1345076, 35, 2, n_days = 140)

import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('tkagg')

plt.plot(list(range(5)))
plt.show()



import numpy as np

class SIR:
    def __init__(self, S: int, I: int, R: int, beta: float,
                 gamma: float, beta_decay: float=0) -> None:
        if beta_decay < 0 or beta_decay > 1:
            raise ValueError('bad beta decay')
        self.S = S
        self.I = I
        self.R = R
        self.N = sum([S,I,R])
        self.beta = beta
        self.gamma = gamma
        self.beta_decay = beta_decay
        self.s = [S]
        self.i = [I]
        self.r = [R]

    def _rectify(self, x: float) -> float:
        out = 0 if x < 0 else x
        return out

    def _Sn(self, S: int, I:int) -> float:
        return self._rectify((-self.beta * S * I) + S)

    def _In(self, S: int, I:int) -> float:
        return self._rectify((self.beta * S * I - self.gamma * I) + I)

    def _Rn(self, I: int, R: int) -> float:
        return self._rectify((self.gamma * I + R))

    def _step(self, S: int, I: int, R: int) -> (float, float, float):
        Sn = self._Sn(S, I)
        Rn = self._Rn(I, R)
        In = self._In(S, I)
        scale = self.N / (Sn + Rn + In)
        S = Sn * scale
        I = In * scale
        R = Rn * scale
        return S, I , R

    def sir(self, n_days: int) -> (np.ndarray, np.ndarray, np.ndarray):
        S = self.S
        I = self.I
        R = self.R
        self.s = [S]
        self.i = [I]
        self.r = [R]
        for day in range(n_days):
            S, I, R = self._step(S, I, R)
            self.beta *= (1-self.beta_decay)
            self.s.append(S)
            self.i.append(I)
            self.r.append(R)
        return (np.asarray(x) for x in [self.s, self.i, self.r])


class SIRD(SIR):
    def __init__(self, N: int, I: int, R: int, D: int, beta: float, gamma: float, mu: float,
                 birth_rate: float, beta_decay: float) -> None:
        S = N - (I+R+D)
        super().__init__(S,I,R,beta,gamma, beta_decay)
        self.N = N
        self.birth_rate = birth_rate
        self.D = D
        self.mu = mu
        self.d = [D]

    def _Sn(self, S: int, I: int) -> float:
        out = (self.birth_rate - self.mu * S - self.beta * S * I) + S
        return self._rectify(out)

    def _In(self, S: int, I: int) -> float:
        lhs = self.beta * S * I
        rhs = self.gamma * I - self.mu * I
        out = lhs - rhs + I
        return self._rectify(out)

    def _Rn(self, I: int, R: int) -> float:
        lhs = self.gamma * I
        rhs = self.mu * R
        out = lhs - rhs + R
        return self._rectify(out)

    def _step(self, S: int, I: int, R: int) -> (float, float, float, float):
        S, I, R = super()._step(S, I, R)
        return S, I, R

    def sir(self, n_days: int) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
        S = self.S
        I = self.I
        R = self.R
        D = self.D
        self.s = [S]
        self.i = [I]
        self.r = [R]
        self.d = [D]
        for day in range(n_days):
            S, I, R = self._step(S, I, R)
            self.d.append(I*self.mu)
            self.beta *= (1-self.beta_decay)
            self.s.append(S)
            self.i.append(I)
            self.r.append(R)
        return (np.asarray(x) for x in [self.s, self.i, self.r, self.d])







import numpy as np
from .sir import SIR, SIRD

def calc_hosp_numbers(i: np.ndarray, rate: float) -> np.ndarray:
    """
    calc_hosp_numbers:
        given an infection curve i, and a rate, how many people will use
        resource XYZ
    Inputs:
        i: np.ndarray = infection curve
        rate: float = rate of whatever
    Outputs: np.ndarray
    """
    return i*rate


def calc_admission_deltas(admits: np.ndarray) -> np.ndarray:
    """
    calc_admission_deltas:
        given admissions to a hosp/icu/ventilators, what is the daily change in
        admissions?
    Inputs:
        admits: np.ndarray = number of admissions over time
    Outputs: np.ndarray
    """
    out = admits[1:] - admits[:-1]
    out[np.where(out < 0)] = 0
    return out


def rolling_sum(a: np.ndarray, window: int) -> np.ndarray:
    """
    rolling_sum:
        calculate the a rolling sum of a numpy array
    Inputs:
        a: np.ndarray = the array you want to calculate a rolling sum for
        window: int = the size of the window to sum for
    Outputs:
        np.ndarray
    """
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    rolled = np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)
    return rolled.sum(-1).copy()


class Penn(SIR):
    """
    Penn:
        Make SIR predictions given the assumed detection rate (out of all infected cases), using
        the parameter estimation method from the Penn model.
    ----------------------------------------------------------------------
    Parameters:
        S: pop size, Number of people affected
        I: infected, Number of positive tests in region
        R: recovered, number of recoveries
        detect_prob: percentage of all infected cases in region which you believe are being detected = 1
        hosp_rate: rate of infected admitted to the hospital = 0.05
        icu_rate: rate of infected who need to be in ICU = 0.02
        vent_rate: rate of infected who need ventilators = 0.01
        contact_reduction: percent contact reduced by social distancing = 0
        t_double: time to double number of infected = 6.
        beta_decay: decay rate of beta, which represents how often a contact results in a new infection = 0
        vent_los: time one patient takes up a ventilator=10
        hos_los: time one patient takes up a normal hospital bed = 7
        icu_los: time one patient takes up an ICU bed = 9
        recover_time: time to get better, to shift from I to R = 14

    Attributes:
        rates: dict = rate of hospitalization, icu, and ventilators
        los: dict = length of stay at hospital, icu, and ventilator
        contact_reduction: float = contact reduction
        t_double: float = number of days to double
        intrinsic_growth: float = growth rate
        recover_time: float = time to recover from illness
        beta: float =  how often a contact results in a new infection
        gamma:  float = rate at which an infected person recovers
        beta_decay:  float = decay rate of beta
        r_naught:  float = speadability of disease
        r_t:  float = r_naught after distancing
        S: int = number of susceptible people
        I: int = number of actually infected people
        R: int = number of recovered people


    Methods:
        sir(n_days):
            run simulation, see docstring for sir
    """
    def __init__(self, S: int, I: int, R: int, detect_prob: float = 1,
                 hosp_rate: float=0.05, icu_rate: float=0.02,
                 vent_rate: float=0.01, contact_reduction: float=0.,
                 t_double: float=6, beta_decay: float=0, vent_los: float=10,
                 hos_los: float=7, icu_los: float=9, recover_time: float=14) -> None:
        self.rates = {'hospital': hosp_rate,'icu': icu_rate, 'ventilator':vent_rate}
        self.los = dict(zip(self.rates.keys(), [hos_los, icu_los, vent_los]))
        self.contact_reduction = contact_reduction
        self.t_double = t_double
        self.intrinsic_growth = 2**(1/t_double) - 1
        self.recover_time = recover_time
        gamma = 1/self.recover_time
        beta = ((self.intrinsic_growth + gamma) / S)  * (1-contact_reduction)
        self.r_t = beta/gamma * S
        self.r_naught = self.r_t / (1-contact_reduction)
        self.I = I / detect_prob
        super().__init__(S, self.I, R, beta, gamma, beta_decay)

    def sir(self, n_days: int) -> (dict, dict):
        """
        sir: fit SIR model to the data
        -----------------------------------
        Inputs:
            n_days: int = number of days ahead to predict
        Outputs:
            curve: dict :
                susceptible: np.ndarray = history of susceptible pop
                infected: np.ndarray = history of infection
                recovered: np.ndarray = history of recovery
            admissino: dict: of occupancy
                hospitalized
                icu
                ventilator
        """
        s, i, r = super().sir(n_days)
        out = {}
        admissions = {}
        out['infected'] = i
        out['recovered'] = r
        out['susceptible'] = s
        for k in self.rates.keys():
            # calculate raw numbers
            out[k] = calc_hosp_numbers(i, self.rates[k])
            # turn to new admissions per day
            admissions[k] = out[k]
            admissions[k] = calc_admission_deltas(admissions[k])
            admissions[k] = rolling_sum(admissions[k], self.los[k])
        return out, admissions


class PennDeath(SIRD):
    """
    PennDeath:
        Make SIR predictions given the number of deaths today
    ----------------------------------------------------------------------
    Parameters:
        N: pop size, Number of people affected
        I: infected, Number of positive tests in region
        R: recovered, number of recoveries
        D: dead, number dead, total
        D_today: dead today, number of people who died today
        hosp_rate: rate of infected admitted to the hospital = 0.15
        icu_rate: rate of infected who need to be in ICU = 0.05
        vent_rate: rate of infected who need ventilators = 0.02
        death_rate: rate of infected who die = 0.01
        contact_reduction: percent contact reduced by social distancing = 0
        t_death: time from infection to death = 20
        t_double: time to double number of infected = 6.
        beta_decay: decay rate of beta, which represents how often a contact results in a new infection = 0
        vent_los: time one patient takes up a ventilator=10
        hos_los: time one patient takes up a normal hospital bed = 7
        icu_los: time one patient takes up an ICU bed = 9
        recover_time: time to get better, to shift from I to R = 20
        birth_rate: Daily rate of birth. Assumed to be 0, as babies are
        hopefully not susceptible

    Attributes:
        rates: dict = rate of hospitalization, icu, and ventilators
        los: dict = length of stay at hospital, icu, and ventilator
        contact_reduction: float = contact reduction
        t_double: float = number of days to double
        intrinsic_growth: float = growth rate
        recover_time: float = time to recover from illness
        beta: float =  how often a contact results in a new infection
        gamma:  float = rate at which an infected person recovers
        beta_decay:  float = decay rate of beta
        r_naught:  float = speadability of disease
        r_t:  float = r_naught after distancing
        S: int = number of susceptible people
        I: int = number of actually infected people
        R: int = number of recovered people
        D: int = number of dead


    Methods:
        sir(n_days):
            run simulation, see docstring for sir
    """
    def __init__(self, N: int, I: int, R: int, D: int, D_today: int,
                 hosp_rate: float=0.15, icu_rate: float=0.05,
                 vent_rate: float=0.02, death_rate: float=0.01,
                 contact_reduction: float=0., t_death: float=20,
                 t_double: float=6, beta_decay: float=0, vent_los: float=10,
                 hos_los: float=7, icu_los: float=9, recover_time: float=20,
                 birth_rate: float=0):
        self.rates = {'hospital': hosp_rate,'icu': icu_rate, 'ventilator':vent_rate}
        self.los = dict(zip(self.rates.keys(), [hos_los, icu_los, vent_los]))
        self.contact_reduction = contact_reduction
        self.t_double = t_double
        self.intrinsic_growth = 2**(1/t_double) - 1
        self.recover_time = recover_time
        S = N - (I + R + D)
        gamma = 1/self.recover_time
        beta = ((self.intrinsic_growth + gamma) / S)  * (1-contact_reduction)
        self.r_t = beta/gamma * S
        self.r_naught = self.r_t / (1-contact_reduction)
        if D_today == 0:
            self.I = I
        else:
            infect_t_death = D_today/death_rate
            penn = Penn(S, I, R, 1,
                 hosp_rate, icu_rate ,
                 vent_rate , contact_reduction,
                 t_double, beta_decay, vent_los,
                 hos_los, icu_los, recover_time)
            self.I = penn.sir(t_death)[0]['infected'][-1]
        super().__init__(N, self.I, R, D,beta, gamma, death_rate, birth_rate, beta_decay)


    def sir(self, n_days: int) -> (dict, dict):
        """
        sir: fit SIR model to the data
        -----------------------------------
        Inputs:
            n_days: int = number of days ahead to predict
        Outputs:
            curve: dict :
                susceptible: np.ndarray = history of susceptible pop
                infected: np.ndarray = history of infection
                recovered: np.ndarray = history of recovery
                dead: np.ndarray = history of death
                hospital: np.ndarray = history of hospitalizations
                icu: np.ndarray = history of icu
                ventilator: np.ndarray = history of ventilator
            occupancy: dict of occupancy:
                hospitaliz
                icu
                ventilator
        """
        s, i, r, d = super().sir(n_days)
        out = {}
        admissions = {}
        out['infected'] = i
        out['recovered'] = r
        out['susceptible'] = s
        out['dead'] = d
        for k in self.rates.keys():
            # calculate raw numbers
            out[k] = calc_hosp_numbers(i, self.rates[k])
            # turn to new admissions per day
            admissions[k] = out[k]
            admissions[k] = calc_admission_deltas(admissions[k])
            admissions[k] = rolling_sum(admissions[k], self.los[k])
        return out, admissions




import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('tkagg')
plt.style.use('seaborn-paper')
from comodels import PennDeath

tx = PennDeath(330460832, 14250, 125, 205, 57, death_rate = 0.01, birth_rate=0, contact_reduction = 0.)

def plot_penn(Pdp: PennDeath, n_days: int) -> None:
    curve, admissions = Pdp.sir(n_days)
    fig, ax = plt.subplots(1,3, figsize=(15,5))
    for k, v in curve.items():
        if k not in tx.rates.keys() :
            ax[0].plot(v, label=k)
            ax[0].legend()
        else:
            ax[1].plot(v, label=k)
            ax[1].legend()
    ax[1].set_title('Hospital Resource Usage')
    ax[0].set_title('SIR curve')
    for k, v in admissions.items():
        ax[2].plot(v, label = k)
        ax[2].legend()
    ax[2].set_title('Additional Resource Usage by day')
    fig.suptitle(f"No social distancing, total deaths = {int(max(curve['dead']))}")
    plt.show()


plot_penn(tx, 120)



import setuptools
def readme():
    with open('README.md') as readme_file:
        return readme_file.read()


setuptools.setup(
    name = "covid-modeling",
    version = "0.0.7",
    description = "COVID-19 Modeling",
    long_description = readme(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/josephsdavid/COVID-modeling",
    maintainer = "David Josephs",
    maintainer_email = "josephsd@smu.edu",
   #$ packages = setuptools.find_packages(exclude = [
    #    "*weights*", "*viz*", "*data*"
   # ]),
    packages = ['comodels'],
    license = 'MIT',
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires = ['numpy']
)



import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('tkagg')
plt.style.use('seaborn-paper')
from comodels import Penn

print(Penn.__doc__)

print(Penn.sir.__doc__)

tx = Penn(28304596, 223, 0, 1/8)
tx_good = Penn(28304596, 223, 0, 1/8, contact_reduction = 0.33)

fig, axs = plt.subplots(3,2)
for k, v in tx.sir(180)[0].items():
    if k in tx.rates.keys():
        axs[1,0].plot(v, label=k)
        axs[1,0].legend()
axs[1,0].set_title('Hospital Resource Usage, No social distancing, TX')
for k, v in tx.sir(180)[0].items():
    if k not in tx.rates.keys():
        axs[0,0].plot(v, label=k)
        axs[0,0].legend()
axs[0,0].set_title('SIR chart, No social distancing, TX')
for k, v in tx.sir(180)[1].items():
        axs[2,0].plot(v, label=k)
        axs[2,0].legend()
axs[2,0].set_title('Daily Hospital admissions, no social distancing, TX')
for k, v in tx_good.sir(180)[0].items():
    if k in tx_good.rates.keys():
        axs[1,1].plot(v, label=k)
        axs[1,1].legend()
axs[1,1].set_title('Hospital Resource Usage, Social contact reduced by 33%, TX')
for k, v in tx_good.sir(180)[0].items():
    if k not in tx_good.rates.keys():
        axs[0,1].plot(v, label=k)
        axs[0,1].legend()
axs[0,1].set_title('SIR chart, Social contact reduced by 33%, TX')
for k, v in tx_good.sir(180)[1].items():
        axs[2,1].plot(v, label=k)
        axs[2,1].legend()
axs[2,1].set_title('Daily Hospital admissions, Social Contact reduced by 33%, TX')
plt.suptitle("Penn model of TX, given sensitivity of test = 1/8")
plt.show()

